import subprocess
import json
from .template import template, template_any
from .jinja_context import pull_jinja_context
from .errors import fail_on_invalid_db_name
from .config import load_config
from .lpass import pull_lastpass_username_password, db_info_from_lpass
from .db_facts_types import DBConfig, DBCLIConfig, DBFacts, DBName
from .db_config import db_config


def db(db_name: DBName, dbcli_config: DBCLIConfig = None) -> DBFacts:
    """Get connection info for specified database."""

    if dbcli_config is None:
        dbcli_config = load_config()

    config = db_config(dbcli_config, db_name)
    if config is None:
        fail_on_invalid_db_name(db_name)

    db_connect_method = config.get('db_connect_method')
    db_info: DBConfig = {
        'exports': {
            'connection_type': 'direct',
        }
    }
    if db_connect_method is not None:
        db_info['db_connect_method'] = db_connect_method

    jinja_context = pull_jinja_context(db_name, config, dbcli_config)
    for k in config.keys():
        if k == 'jinja_context_name':
            pass
        elif k == 'exports':
            db_info['exports'].update({
                template(exportname,
                         jinja_context): template_any(config['exports'][exportname], jinja_context)
                for exportname in config['exports']
            })
        else:
            db_info[k] = template(config[k], jinja_context)

    if 'db_connect_method' in db_info:
        method = db_info.get('db_connect_method')
        if method is None:
            # nothing to do here - all the info we need is already in
            # dbcli.yml, either fixed or interpolated through the
            # jinja context
            pass
        if 'json_script' in dbcli_config['db_connect_method'][method]:
            json_call_args = \
                dbcli_config['db_connect_method'][method]['json_script']
            json_call =\
                [template(s, (db_info, {})) for s in json_call_args]
            json_str = subprocess.check_output(json_call).decode('utf-8')
            additional_attributes = json.loads(json_str)
            db_info['exports'].update(additional_attributes)
        elif 'pull_lastpass_from' in dbcli_config['db_connect_method'][method]:
            template_for_lastpass_entry_name =\
                dbcli_config['db_connect_method'][method]['pull_lastpass_from']
            lastpass_entry_name = template(template_for_lastpass_entry_name,
                                           (db_info, {}))
            additional_attributes = \
                db_info_from_lpass(lastpass_entry_name)
            db_info['exports'].update(additional_attributes)
        elif 'pull_lastpass_username_password_from' in \
             dbcli_config['db_connect_method'][method]:
            method = dbcli_config['db_connect_method'][method]
            template_for_lastpass_entry_name =\
                method['pull_lastpass_username_password_from']
            lastpass_entry_name = template(template_for_lastpass_entry_name,
                                           (db_info, {}))
            additional_attributes = \
                pull_lastpass_username_password(lastpass_entry_name)
            db_info['exports'].update(additional_attributes)
        else:
            raise SyntaxError(f'Did not understand db_connect_method {method}')

    return db_info['exports']



from __future__ import print_function
import sys
from .db_info import db
from .exports import print_exports
from .list_db_names import list_db_names
from .config_yaml import config_yaml
from typing import List
from .errors import UserErrorException
import argparse
import json


class Runner():
    def __init__(self) -> None:
        pass

    def dump_json(self, args: argparse.Namespace) -> None:
        db_name_str: str = args.dbname[0]
        db_name = db_name_str.split('-')

        db_info = db(db_name)

        print(json.dumps(db_info, sort_keys=True, indent=4))

    def dump_config(self, args: argparse.Namespace):
        db_name_str: str = args.dbname[0]
        db_name = db_name_str.split('-')

        db_info = db(db_name)
        print(config_yaml(db_name_str, db_info), end='')

    def dump_sh(self, args: argparse.Namespace):
        db_name_str: str = args.dbname[0]
        db_name = db_name_str.split('-')

        db_info = db(db_name)
        print_exports(db_info)

    def run(self, argv: List[str]) -> int:
        try:
            desc = 'Pull information about databases from user-friendly names'
            parser = argparse.ArgumentParser(prog='db-facts', description=desc)
            subparsers = parser.add_subparsers()

            list_parser = subparsers.add_parser('list', help='List available dbnames')
            list_parser.set_defaults(func=lambda args: list_db_names())

            json_parser = subparsers.add_parser('json', help='Report output in JSON format')
            json_parser.add_argument('dbname', nargs=1,
                                     help=('Friendly name of database '
                                           '(e.g., "redshift", "dmv", '
                                           '"abc-dev-dbadmin")'))
            json_parser.set_defaults(func=self.dump_json)

            config_parser = subparsers.add_parser('config',
                                                  help='Report output in db-facts config format')
            config_parser.add_argument('dbname', nargs=1,
                                       help=('Friendly name of database '
                                             '(e.g., "redshift", "dmv", '
                                             '"abc-dev-dbadmin")'))
            config_parser.set_defaults(func=self.dump_config)

            sh_parser = subparsers.add_parser('sh',
                                              help=('Report output in Bourne shell envionment '
                                                    'variable format '))
            sh_parser.add_argument('dbname', nargs=1,
                                   help=('Friendly name of database '
                                         '(e.g., "redshift", "dmv", '
                                         '"abc-dev-dbadmin")'))
            sh_parser.set_defaults(func=self.dump_sh)

            args = parser.parse_args(argv[1:])
            if 'func' not in args:
                parser.print_help(file=sys.stderr)
                return 1
            args.func(args)

            return 0
        except UserErrorException as e:
            print(str(e), file=sys.stderr)
            return 1



import pipes

formatted_key = {
    'password': 'DB_PASSWORD',
    'host': 'DB_HOST',
    'user': 'DB_USERNAME',
    'type': 'DB_TYPE',
    'protocol': 'DB_PROTOCOL',
    'port': 'DB_PORT',
    'database': 'DB_DATABASE',
}


def print_exports(exports):
    for k, v in sorted(exports.items()):
        env_var = formatted_key.get(k, k.upper())
        print("export " + env_var)
        print(env_var + "=" + pipes.quote(str(v)))




from db_facts.runner import Runner
import unittest
from io import StringIO
from unittest.mock import patch
import json


@patch('db_facts.runner.db')
@patch('sys.stderr', new_callable=StringIO)
@patch('sys.stdout', new_callable=StringIO)
class TestRunnerSuccessJson(unittest.TestCase):
    def test_runner_lpass_json(self,
                               mock_stdout,
                               mock_stderr,
                               mock_db):
        runner = Runner()
        mock_db.return_value = {
            'password': 'password',
            'host': 'host',
            'user': 'user',
            'type': 'type',
            'protocol': 'protocol',
            'port': 123,
            'database': 'dbname',
            'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
            'connection_type': 'connection_type',
        }

        self.assertEqual(0, runner.run(['/bin/db-facts', 'json', 'foo']))
        mock_db.assert_called_with(['foo'])
        self.assertEqual(mock_stderr.getvalue(), '')
        parsed_json_value = {
            'database': 'dbname',
            'user': 'user',
            'password': 'password',
            'port': 123,
            'protocol': 'protocol',
            'type': 'type',
            'host': 'host',
            'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
            'connection_type': 'connection_type'
        }
        self.assertEqual(json.loads(mock_stdout.getvalue()),
                         parsed_json_value)



import base64
from .db_facts_types import DBName, DBCLIConfig, JinjaContext, JinjaFilters
from typing import Tuple


def pull_base64_jinja_context(db_name: DBName,
                              dbcli_config: DBCLIConfig) -> Tuple[JinjaContext,
                                                                  JinjaFilters]:
    return ({},
            {
                'b64decode': lambda s: base64.b64decode(s).decode('utf-8'),
                'b64encode': lambda s: base64.b64encode(s.encode('utf-8')).decode('utf-8'),
            })



from db_facts.db_info import db
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.db_info.pull_jinja_context')
@patch('db_facts.db_info.db_info_from_lpass')
@patch('db_facts.lpass.lpass_field')
class TestDBInfo(unittest.TestCase):
    def test_db_info_two_level_with_config(self,
                                           mock_lpass_field,
                                           mock_db_info_from_lpass,
                                           mock_pull_jinja_context):

        self.maxDiff = True

        expected_result = {
            'database': 'database',
            'host': 'host',
            'password': 'password',
            'port': 'port',
            'protocol': 'protocol',
            'type': 'type',
            'user': 'user',
            'connection_type': 'direct'
        }
        mock_db_info_from_lpass.return_value = {
            'password': 'password',
            'host': 'host',
            'user': 'user',
            'type': 'type',
            'protocol': 'protocol',
            'port': 'port',
            'database': 'database',
        }
        mock_jinja_context = {}
        mock_filters = {}
        mock_pull_jinja_context.return_value = (mock_jinja_context,
                                                mock_filters)
        db_facts = db(['bazzle', 'bing'], dbcli_config=mock_dbcli_config)

        self.assertEqual(expected_result, db_facts)
        mock_pull_jinja_context.\
            assert_called_with(['bazzle', 'bing'],
                               mock_dbcli_config['dbs']['bazzle-bing'],
                               mock_dbcli_config)

    def test_db_info_two_level_with_one_level_config(self,
                                                     mock_lpass_field,
                                                     mock_db_info_from_lpass,
                                                     mock_pull_jinja_context):

        self.maxDiff = None

        expected_result = {
            'database': 'database',
            'host': 'host',
            'password': 'password',
            'port': 'port',
            'protocol': 'protocol',
            'type': 'type',
            'user': 'user',
            'connection_type': 'direct'
        }
        mock_db_info_from_lpass.return_value = {
            'password': 'password',
            'host': 'host',
            'user': 'user',
            'type': 'type',
            'protocol': 'protocol',
            'port': 'port',
            'database': 'database',
        }
        mock_pull_jinja_context.return_value = ({}, {})
        db_facts = db(['frazzle', 'mumble'], dbcli_config=mock_dbcli_config)
        self.assertEqual(expected_result, db_facts)
        mock_pull_jinja_context.\
            assert_called_with(['frazzle', 'mumble'],
                               mock_dbcli_config['dbs']['frazzle'],
                               mock_dbcli_config)



from db_facts.db_info import db
from db_facts.errors import UserErrorException
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.db_info.pull_jinja_context')
@patch('db_facts.db_info.subprocess')
class TestDBInfoErrors(unittest.TestCase):
    def test_db_info_bad_db(self,
                            mock_subprocess,
                            mock_pull_jinja_context):
        mock_pull_jinja_context.return_value = ({}, {})
        with self.assertRaises(UserErrorException):
            db(['not-there'], dbcli_config=mock_dbcli_config)

    def test_db_info_invalid_db_connect_method(self,
                                               mock_subprocess,
                                               mock_pull_jinja_context):
        mock_pull_jinja_context.return_value = ({}, {})
        with self.assertRaises(SyntaxError):
            db(['bing'], dbcli_config=mock_dbcli_config)



from pathlib import Path
from nose.tools import eq_
from db_facts.config import merge_configs, load_config


def test_merge_configs():
    result = merge_configs(
        {'a': {'b': 1, 'c': 2}},
        {'a': {'b': 2, 'd': 3}, 'e': {'f': 'g'}},
        {'e': {'f': 'h'}},
    )

    eq_(result, {
        'a': {'b': 2, 'c': 2, 'd': 3},
        'e': {'f': 'h'},
    })


def test_load_config():
    """Tests that defaults are preserved if override files don't include an override."""

    config = load_config(config_paths_to_load=[Path(__file__).parent / 'files' / 'test.yml'])
    eq_('blah', config['dbs']['newconfig']['exports']['host'])


def test_load_config_no_paths():
    """Tests that can disable default config."""
    config = load_config(config_paths_to_load=[])
    eq_(config, {})



from pathlib import Path
from unittest.mock import patch
from nose.tools import eq_
from db_facts.config import find_configs


def fake_enumerate_configs(config_prefix, file_prefix):
    if config_prefix == Path('/etc'):
        return [config_prefix / f'{file_prefix}dbfacts.d' / '001_fake.yml']
    elif config_prefix == Path('/usr/local/etc'):
        return [config_prefix / f'{file_prefix}dbfacts.d' / '002_fake.yml']
    elif config_prefix == Path.home():
        return [config_prefix / f'{file_prefix}dbfacts.d' / '003_fake.yml']
    elif config_prefix == Path('.'):
        return [config_prefix / f'{file_prefix}dbfacts.d' / '004_fake.yml']
    else:
        return []


@patch('db_facts.config.enumerate_configs', new=fake_enumerate_configs)
@patch('db_facts.config.Path.is_file')
def test_find_configs(mock_is_file):
    mock_is_file.return_value = True
    eq_(find_configs(), [
        Path('/etc/dbfacts.yml'),
        Path('/etc/dbfacts.d/001_fake.yml'),
        Path('/usr/local/etc/dbfacts.yml'),
        Path('/usr/local/etc/dbfacts.d/002_fake.yml'),
        Path.home() / '.dbfacts.yml',
        Path.home() / '.dbfacts.d/003_fake.yml',
        Path('.dbfacts.yml'),
        Path('.dbfacts.d/004_fake.yml')
    ])


@patch.dict('os.environ', {
    'DB_FACTS_PATH': '/foo/bar/dbfacts.yml:/foo/bar/custom.yml',
})
@patch('db_facts.config.enumerate_configs', new=fake_enumerate_configs)
@patch('db_facts.config.Path.is_file')
def test_find_configs_with_env_var(mock_is_file):
    mock_is_file.return_value = True
    eq_(find_configs(), [
        Path('/etc/dbfacts.yml'),
        Path('/etc/dbfacts.d/001_fake.yml'),
        Path('/usr/local/etc/dbfacts.yml'),
        Path('/usr/local/etc/dbfacts.d/002_fake.yml'),
        Path.home() / '.dbfacts.yml',
        Path.home() / '.dbfacts.d/003_fake.yml',
        Path('.dbfacts.yml'),
        Path('.dbfacts.d/004_fake.yml'),
        Path('/foo/bar/custom.yml'),
        Path('/foo/bar/dbfacts.yml'),
    ])



import os
from .db_facts_types import JinjaContext, DBName, DBCLIConfig


def pull_env_jinja_context(db_name: DBName, dbcli_config: DBCLIConfig) -> JinjaContext:
    return {
        # Returns None if key not found unless default specified
        'getenv': os.getenv,
        # Raises if key not found
        'env': lambda key: os.environ[key],
    }



from db_facts.db_info import db
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.db_info.pull_jinja_context')
@patch('db_facts.db_info.subprocess')
class TestDBInfoJsonScript(unittest.TestCase):
    def test_db_info_json_script(self,
                                 mock_subprocess,
                                 mock_pull_jinja_context):

        expected_result = {
            'a': 'b',
            'c': 'd',
            'connection_type': 'direct',
        }
        mock_subprocess.check_output.return_value = b'{"a": "b", "c": "d"}'
        mock_pull_jinja_context.return_value = ({}, {})
        db_facts = db(['baz'], dbcli_config=mock_dbcli_config)
        self.assertEqual(expected_result, db_facts)
        mock_pull_jinja_context.\
            assert_called_with(['baz'],
                               mock_dbcli_config['dbs']['baz'],
                               mock_dbcli_config)
        mock_subprocess.check_output.\
            assert_called_with(['some-custom-json-script'])



import logging
from pathlib import Path
import os
from os.path import expanduser
import yaml


logger = logging.getLogger(__name__)

# Prefixes to scan for config files (from lowest priority to highest priority)
CONFIG_PREFIXES = ['/etc', '/usr/local/etc', '~', '.']


def merge_configs(*configs):
    """
    Merges dictionaries of dictionaries, by combining top-level dictionaries with last value taking
    precedence.

    For example:
    >>> merge_configs({'a': {'b': 1, 'c': 2}}, {'a': {'b': 2, 'd': 3}})
    {'a': {'b': 2, 'c': 2, 'd': 3}}
    """
    merged_config = {}
    for config in configs:
        for k, v in config.items():
            if k in merged_config:
                merged_config[k].update(v)
            else:
                merged_config[k] = v.copy()
    return merged_config


def enumerate_configs(config_prefix, file_prefix):
    return (config_prefix / f'{file_prefix}dbfacts.d').glob('*')


def find_configs():
    config_paths_to_load = []
    for config_prefix in map(lambda p: Path(expanduser(p)), CONFIG_PREFIXES):
        if config_prefix in [Path.home(), Path('.')]:
            file_prefix = '.'
        else:
            file_prefix = ''

        config_paths_to_load.append(config_prefix / f'{file_prefix}dbfacts.yml')
        config_paths_to_load.extend(enumerate_configs(config_prefix, file_prefix))
    if 'DB_FACTS_PATH' in os.environ:
        paths = [
            Path(expanduser(rawpath))
            for rawpath in os.environ['DB_FACTS_PATH'].split(':')
        ]
        config_paths_to_load.extend(reversed(paths))

    return [path for path in config_paths_to_load if path.is_file()]


def load_config(config_paths_to_load=None):
    loaded_configs = []
    if config_paths_to_load is None:
        config_paths_to_load = find_configs()

    for config_path in config_paths_to_load:
        path = Path(config_path)
        logger.info('Merging db facts config from %s', config_path)
        with path.open('r') as configfile:
            config = yaml.safe_load(configfile)
            if isinstance(config, dict):
                loaded_configs.append(config)
            else:
                logger.warning('Config %s appears to be invalid', config_path)

    return merge_configs(*loaded_configs)



from db_facts.runner import Runner
import unittest
from io import StringIO
from unittest.mock import patch


@patch('sys.stderr', new_callable=StringIO)
@patch('sys.stdout', new_callable=StringIO)
class TestRunnerList(unittest.TestCase):

    @patch('db_facts.list_db_names.load_config')
    def test_runner_list(self,
                         mock_load_config,
                         mock_stdout,
                         mock_stderr):
        runner = Runner()
        mock_load_config.return_value = {
            'dbs': {
                'mydb1': {
                    'description': 'My favorite database',
                    'password': 'password',
                    'host': 'host',
                    'user': 'user',
                    'type': 'type',
                    'protocol': 'protocol',
                    'port': 123,
                    'database': 'dbname',
                    'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
                    'connection_type': 'connection_type',
                },
                'mydb2': {
                    'password': 'password',
                    'host': 'host',
                    'user': 'user',
                    'type': 'type',
                    'protocol': 'protocol',
                    'port': 123,
                    'database': 'dbname',
                    'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
                    'connection_type': 'connection_type',
                }
            }
        }

        self.assertEqual(0, runner.run(['/bin/db-facts', 'list']))
        mock_load_config.assert_called_with()
        self.assertEqual(mock_stderr.getvalue(), '')
        self.assertEqual(mock_stdout.getvalue(),
                         ("Available db_names:\n"
                          "* mydb1 (My favorite database)\n"
                          "* mydb2\n"))



from db_facts.jinja_context import pull_jinja_context
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.util.subprocess')
class TestNilJinjaContext(unittest.TestCase):
    def test_nil_context(self, mock_subprocess):
        config = {}
        context, filters = pull_jinja_context(['notthere', 'alsonotthere'],
                                              config, mock_dbcli_config)
        self.assertEquals(context, {})



from db_facts.runner import Runner
import unittest
from io import StringIO
from unittest.mock import patch


@patch('db_facts.runner.db')
@patch('sys.stderr', new_callable=StringIO)
@patch('sys.stdout', new_callable=StringIO)
class TestRunner(unittest.TestCase):
    def test_runner_lpass(self,
                          mock_stdout,
                          mock_stderr,
                          mock_db):
        runner = Runner()
        mock_db.return_value = {
            'password': 'password',
            'host': 'host',
            'user': 'user',
            'type': 'type',
            'protocol': 'protocol',
            'port': 123,
            'database': 'dbname',
            'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
            'connection_type': 'connection_type',
        }

        self.assertEqual(0, runner.run(['/bin/db-facts', 'sh', 'foo']))
        mock_db.assert_called_with(['foo'])
        self.assertEqual(mock_stderr.getvalue(), '')
        self.assertEqual(mock_stdout.getvalue(),
                         'export CONNECTION_TYPE\n'
                         'CONNECTION_TYPE=connection_type\n'
                         'export DB_DATABASE\n'
                         'DB_DATABASE=dbname\n'
                         'export DB_HOST\n'
                         'DB_HOST=host\n'
                         'export LASTPASS_SHARE_NAME_SUFFIX\n'
                         'LASTPASS_SHARE_NAME_SUFFIX='
                         'lastpass_share_name_suffix\n'
                         'export DB_PASSWORD\n'
                         'DB_PASSWORD=password\n'
                         'export DB_PORT\n'
                         'DB_PORT=123\n'
                         'export DB_PROTOCOL\n'
                         'DB_PROTOCOL=protocol\n'
                         'export DB_TYPE\n'
                         'DB_TYPE=type\n'
                         'export DB_USERNAME\n'
                         'DB_USERNAME=user\n')



from typing import Any, Dict, Callable, Union, Tuple, List, TYPE_CHECKING
# note: this file cannot be named 'types.py' or I encounter IDE heartache:
#
#  https://stackoverflow.com/questions/36669895/site-py-attributeerror-module-object-has-no-attribute-moduletype-upon-runn
#
# 🤷


if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    # Declare this only during type checking, because I don't want to
    # require mypy as a dependency (the types are built into the
    # language, the type checker and its extensions are a separate
    # package).

    # As you encounter needs to add fields below for other types of
    # databases, feel free to PR additions here (or maintain your own
    # type/ignore this one, of course).
    class DBFacts(TypedDict, total=False):
        user: str
        password: str
        host: str
        type: str
        protocol: str
        port: int
        database: str
        bq_default_project_id: str
        bq_default_dataset_id: str
        bq_service_account_json: str
else:
    DBFacts = Dict[str, Any]
DBName = List[str]
LastPassUsernamePassword = Dict[str, str]
DBConfig = Dict[str, Any]
DBCLIConfig = Any
JinjaContext = Dict[str, Any]
JinjaFilter = Callable[[Any], Any]
JinjaFilters = Dict[str, JinjaFilter]
JinjaContextPuller = Callable[[DBName, DBCLIConfig],
                              Union[JinjaContext, Tuple[JinjaContext, JinjaFilters]]]



from db_facts.jinja_context import pull_jinja_context
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.util.subprocess')
class TestMultiJinjaContext(unittest.TestCase):
    def test_multi(self, mock_subprocess):
        config = {
            'jinja_context_name': [
                'env',
                'base64',
            ]
        }
        mock_subprocess.check_output.return_value = b'username'
        context, filters = pull_jinja_context(['gaggle'], config, mock_dbcli_config)
        decode = filters['b64decode']
        encode = filters['b64encode']
        self.assertEqual(decode(encode('foo')), 'foo')



import unittest
from unittest.mock import patch, Mock
from db_facts.jinja_context import get_context_pullers
import db_facts


class TestJinjaContext(unittest.TestCase):
    def setUp(self):
        db_facts.jinja_context._context_pullers = None

    def tearDown(self):
        db_facts.jinja_context._context_pullers = None

    @patch('db_facts.jinja_context.pkg_resources')
    def test_get_context_pullers(self, mock_pkg_resources):
        mock_entry_point_a = Mock(name='entry_point_a')
        mock_pkg_resources.iter_entry_points.return_value = [mock_entry_point_a]
        pullers = get_context_pullers()
        self.assertEqual(pullers, {
            mock_entry_point_a.name:
            mock_entry_point_a.load.return_value
        })



import jinja2
from typing import Any, Tuple
from .db_facts_types import JinjaContext, JinjaFilters


def template(s: str, jinja_context_and_filters: Tuple[JinjaContext, JinjaFilters]) -> str:
    jinja_context, jinja_filters = jinja_context_and_filters
    environment = jinja2.Environment()
    environment.filters.update(jinja_filters)
    template = environment.from_string(s)
    return template.render(jinja_context)


def template_any(o: Any, jinja_context_and_filters: Tuple[JinjaContext, JinjaFilters]):
    if isinstance(o, str):
        return template(o, jinja_context_and_filters)
    else:
        return o



from db_facts.runner import Runner
import unittest
from io import StringIO
from unittest.mock import patch
import yaml


@patch('db_facts.runner.db')
@patch('sys.stderr', new_callable=StringIO)
@patch('sys.stdout', new_callable=StringIO)
class TestRunnerSuccessConfig(unittest.TestCase):

    def test_runner_lpass_config(self,
                                 mock_stdout,
                                 mock_stderr,
                                 mock_db):
        runner = Runner()
        mock_db.return_value = {
            'password': 'password',
            'host': 'host',
            'user': 'user',
            'type': 'type',
            'protocol': 'protocol',
            'port': 123,
            'database': 'dbname',
            'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
            'connection_type': 'connection_type',
        }

        self.assertEqual(0, runner.run(['/bin/db-facts', 'config', 'foo']))
        mock_db.assert_called_with(['foo'])
        self.assertEqual(mock_stderr.getvalue(), '')
        parsed_yaml_value = {
            'dbs': {
                'foo': {
                    'exports': {
                        'database': 'dbname',
                        'user': 'user',
                        'password': 'password',
                        'port': 123,
                        'protocol': 'protocol',
                        'type': 'type',
                        'host': 'host',
                        'lastpass_share_name_suffix': 'lastpass_share_name_suffix',
                        'connection_type': 'connection_type'
                    }
                }
            }
        }
        self.assertEqual(yaml.load(mock_stdout.getvalue()),
                         parsed_yaml_value)



import unittest
from unittest.mock import patch
from db_facts import lpass


class TestLPass(unittest.TestCase):
    @patch('db_facts.lpass.check_output')
    def test_lpass_field_notes(self, mock_check_output):
        mock_check_output.return_value = "fakenotes\n".encode("utf-8")
        out = lpass.lpass_field('my_name', 'notes')
        mock_check_output.\
            assert_called_with(['lpass', 'show', '--notes', 'my_name'])
        assert out == "fakenotes"

    @patch('db_facts.lpass.check_output')
    def test_lpass_field_username(self, mock_check_output):
        mock_check_output.return_value = "fakeuser\n".encode("utf-8")
        out = lpass.lpass_field('my_name', 'username')
        mock_check_output.\
            assert_called_with(['lpass', 'show', '--username', 'my_name'])
        assert out == "fakeuser"

    @patch('db_facts.lpass.check_output')
    def test_lpass_field_password(self, mock_check_output):
        mock_check_output.return_value = "fakepassword\n".encode("utf-8")
        out = lpass.lpass_field('my_name', 'password')
        mock_check_output.\
            assert_called_with(['lpass', 'show', '--password', 'my_name'])
        assert out == "fakepassword"

    @patch('db_facts.lpass.check_output')
    def test_lpass_field_url(self, mock_check_output):
        mock_check_output.return_value = "fakeurl\n".encode("utf-8")
        out = lpass.lpass_field('my_name', 'url')
        mock_check_output.\
            assert_called_with(['lpass', 'show', '--url', 'my_name'])
        assert out == "fakeurl"

    @patch('db_facts.lpass.check_output')
    def test_lpass_field_field1(self, mock_check_output):
        mock_check_output.return_value = "fakefield1\n".encode("utf-8")
        out = lpass.lpass_field('my_name', 'field1')
        mock_check_output.\
            assert_called_with(['lpass', 'show', '--field=field1', 'my_name'])
        assert out == "fakefield1"

    @patch('db_facts.lpass.check_output')
    def test_db_info_from_lpass(self, mock_check_output):
        def fake_check_output(args):
            assert args[0] == 'lpass'
            assert args[1] == 'show'
            assert args[3] == 'my_lpass_name'
            ret = {
                "--username": 'fakeuser',
                "--password": 'fakepassword',
                "--field=Hostname": 'fakehost',
                "--field=Port": '123',
                "--field=Type": 'faketype',
                "--field=Database": 'fakedatabase',
            }
            return (ret[args[2]] + "\n").encode('utf-8')
        mock_check_output.side_effect = fake_check_output
        db_info = lpass.db_info_from_lpass('my_lpass_name')
        expected_db_info = {
            'database': 'fakedatabase',
            'host': 'fakehost',
            'password': 'fakepassword',
            'port': 123,
            'type': 'faketype',
            'user': 'fakeuser',
            'protocol': 'faketype',  # if we don't know, just pass through
        }
        assert db_info == expected_db_info



from db_facts.jinja_context import pull_jinja_context
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.env_jinja_context.os')
class TestEnvJinjaContext(unittest.TestCase):
    def test_with_env_set(self, mock_os):
        config = {
            'jinja_context_name': 'env',
        }
        context, filters = pull_jinja_context(['whatever', 'whatever'], config, mock_dbcli_config)
        self.assertEqual(context['getenv'], mock_os.getenv)



from db_facts.runner import Runner
import unittest
from io import StringIO
from unittest.mock import patch
from db_facts.errors import UserErrorException


def without_whitespace(s: str) -> str:
    return s.replace(" ", "").replace("\n", "")


@patch('db_facts.runner.db')
@patch('sys.stderr', new_callable=StringIO)
@patch('sys.stdout', new_callable=StringIO)
class TestRunner(unittest.TestCase):
    maxDiff = None

    def test_runner_no_arg(self,
                           mock_stdout,
                           mock_stderr,
                           mock_db):
        runner = Runner()
        out = runner.run(['/bin/db-facts'])
        self.assertEqual(out, 1)
        self.assertIn(without_whitespace('Pull information about databases from'),
                      without_whitespace(mock_stderr.getvalue()))
        self.assertEqual(mock_stdout.getvalue(), '')

    def test_runner_exception(self,
                              mock_stdout,
                              mock_stderr,
                              mock_db):
        runner = Runner()
        mock_db.side_effect = UserErrorException('error message here')

        self.assertEqual(1, runner.run(['/bin/db-facts', 'sh', 'foo']))
        mock_db.assert_called_with(['foo'])
        self.assertEqual(mock_stderr.getvalue(), 'error message here\n')
        self.assertEqual(mock_stdout.getvalue(), '')

    def test_runner_help(self,
                         mock_stdout,
                         mock_stderr,
                         mock_db):
        runner = Runner()
        with self.assertRaises(SystemExit):
            runner.run(['/bin/db-facts', '--help'])
        self.assertEqual(mock_stderr.getvalue(), '')
        helpstr = """
usage: db-facts [-h] {list,json,config,sh} ...

Pull information about databases from user-friendly names

positional arguments:
  {list,json,config,sh}
    list                List available dbnames
    json                Report output in JSON format
    config              Report output in db-facts config format
    sh                  Report output in Bourne shell envionment variable format

optional arguments:
  -h, --help            show this help message and exit"""

        self.assertEqual(without_whitespace(mock_stdout.getvalue()),
                         without_whitespace(helpstr))



from typing import NoReturn
from .db_facts_types import DBName


class UserErrorException(Exception):
    pass


def fail_on_invalid_db_name(db_name: DBName) -> NoReturn:
    raise UserErrorException('-'.join(db_name) + ' is not a valid DB name.  '
                             'To list valid databases, run "db-facts list" and to configure a '
                             'new database, please see '
                             'https://github.com/bluelabsio/db-facts/blob/master/CONFIGURATION.md')



from .db_facts_types import DBConfig, DBCLIConfig, DBName
from typing import Optional


def db_config(dbcli_config: DBCLIConfig, db_name: DBName) -> Optional[DBConfig]:
    dbs = dbcli_config['dbs']

    config_name = "-".join(db_name)

    if config_name in dbs:
        return dbs[config_name]

    if len(db_name) <= 1:
        return None

    # try again without the last component - Sometimes we want the
    # second and third parts of a db_name to be parsed dynamically by
    # the jinja context functions, and sometimes that's overkill and
    # we can just specify the config for specific two- or three- entry
    # db_names.
    return db_config(dbcli_config, db_name[0:-1])



from db_facts.db_info import db
import unittest
from unittest.mock import patch
from .mock_dbcli_config import mock_dbcli_config


@patch('db_facts.db_info.pull_jinja_context')
@patch('db_facts.db_info.db_info_from_lpass')
@patch('db_facts.lpass.lpass_field')
class TestDBInfoLastPass(unittest.TestCase):
    def test_db_info_lastpass(self,
                              mock_lpass_field,
                              mock_db_info_from_lpass,
                              mock_pull_jinja_context):

        expected_result = {
            'database': 'database',
            'host': 'host',
            'password': 'password',
            'port': 'port',
            'protocol': 'protocol',
            'type': 'type',
            'user': 'user',
            'connection_type': 'direct'
        }
        mock_db_info_from_lpass.return_value = {
            'password': 'password',
            'host': 'host',
            'user': 'user',
            'type': 'type',
            'protocol': 'protocol',
            'port': 'port',
            'database': 'database',
        }
        mock_pull_jinja_context.return_value = ({}, {})
        db_facts = db(['bazzle'], dbcli_config=mock_dbcli_config)
        mock_db_info_from_lpass.assert_called_with('lpass entry name')
        self.assertEqual(expected_result, db_facts)
        mock_pull_jinja_context.\
            assert_called_with(['bazzle'],
                               mock_dbcli_config['dbs']['bazzle'],
                               mock_dbcli_config)

    def test_db_info_pull_lastpass_user_and_pass_only(self,
                                                      mock_lpass_field,
                                                      mock_db_info_from_lpass,
                                                      mock_pull_jinja_context):

        lpass_entry = {
            'username': 'user',
            'password': 'password',
        }

        def fake_lpass_field(entry_name, field_name):
            assert entry_name == 'lpass entry name'
            return lpass_entry[field_name]
        mock_lpass_field.side_effect = fake_lpass_field
        expected_result = {
            'password': 'password',
            'user': 'user',
            'connection_type': 'direct',
            'some_additional': 'export',
            'a_numbered_export': 123,
        }
        mock_pull_jinja_context.return_value = ({}, {})
        db_facts = db(['frink'], dbcli_config=mock_dbcli_config)
        mock_lpass_field.assert_called_with('lpass entry name', 'password')
        self.assertEqual(expected_result, db_facts)
        mock_pull_jinja_context.\
            assert_called_with(['frink'],
                               mock_dbcli_config['dbs']['frink'],
                               mock_dbcli_config)



from .db_facts_types import DBFacts
import yaml


def config_yaml(db_name: str, db_facts: DBFacts) -> str:
    config = {
        'dbs': {
            db_name: {
                'exports': db_facts
            },
        },
    }
    return yaml.dump(config)



from .db_facts_types import (DBName, DBConfig, DBCLIConfig, JinjaContext, JinjaFilters,
                             JinjaContextPuller)
from typing import Dict, Tuple, Optional
import pkg_resources
import logging


logger = logging.getLogger(__name__)


_context_pullers: Optional[Dict[str, JinjaContextPuller]] = None


def get_context_pullers() -> Dict[str, JinjaContextPuller]:
    global _context_pullers

    if _context_pullers is not None:
        return _context_pullers

    # https://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins
    # https://packaging.python.org/guides/creating-and-discovering-plugins/#using-naming-convention
    _context_pullers = {
        entry_point.name: entry_point.load()
        for entry_point
        in pkg_resources.iter_entry_points('db_facts.jinja_contexts')
    }

    return _context_pullers


def pull_jinja_context(db_name: DBName,
                       config: DBConfig,
                       dbcli_config: DBCLIConfig) -> Tuple[JinjaContext, JinjaFilters]:
    if 'jinja_context_name' not in config:
        return ({}, {})
    else:
        all_jinja_context_names = config['jinja_context_name']
        if not isinstance(all_jinja_context_names, list):
            all_jinja_context_names = [all_jinja_context_names]

        full_context: JinjaContext = {}
        full_filters: JinjaFilters = {}
        for jinja_context_name in all_jinja_context_names:
            pullers = get_context_pullers()
            if jinja_context_name in pullers:
                context_or_context_and_filters = \
                    pullers[jinja_context_name](db_name, dbcli_config)
                context: JinjaContext
                filters: Optional[JinjaFilters] = None
                if isinstance(context_or_context_and_filters, tuple):
                    context, filters = context_or_context_and_filters
                else:
                    context = context_or_context_and_filters
                full_context.update(context)
                if filters is not None:
                    full_filters.update(filters)
            else:
                raise Exception('Invalid jinja_context_name (' +
                                jinja_context_name + ')')

        return full_context, full_filters



mock_dbcli_config = {
    'db_connect_method': {
        'lpass': {
            'pull_lastpass_from': "{{ lastpass_entry }}",
        },
        'lpass_user_and_pass_only': {
            'pull_lastpass_username_password_from': "{{ lastpass_entry }}",
        },
        'my-json-script': {
            'json_script': [
                'some-custom-json-script'
            ]
        },
        'invalid-method': {
        },
    },
    'dbs': {
        'baz': {
            'db_connect_method': 'my-json-script',
        },
        'bing': {
            'db_connect_method': 'invalid-method',
        },
        'bazzle': {
            'db_connect_method': 'lpass',
            'lastpass_entry': 'lpass entry name'
        },
        'bazzle-bing': {
            'db_connect_method': 'lpass',
            'lastpass_entry': 'different lpass entry name'
        },
        'frazzle': {
            'db_connect_method': 'lpass',
            'lastpass_entry': 'lpass entry name'
        },
        'frink': {
            'db_connect_method': 'lpass_user_and_pass_only',
            'lastpass_entry': 'lpass entry name',
            'jinja_context_name': 'standard',
            'exports': {
                'some_additional': 'export',
                'a_numbered_export': 123
            },
        },
        'gaggle': {
            'jinja_context_name': [
                'env',
                'base64',
            ],
            'exports': {
                'type': 'bigquery',
                'protocol': 'bigquery',
                'bq_account': 'bq_itest',
                'bq_service_account_json':
                "{{ env('ITEST_BIGQUERY_SERVICE_ACCOUNT_JSON_BASE64') | b64decode }}",
                'bq_default_project_id': 'bluelabs-tools-dev',
                'bq_default_dataset_id': 'bq_itest',
            },
        },
    },
    'orgs': {
        'myorg': {
            'full_name': 'MyOrg',
        },
    },
}



#!/usr/bin/env python3
"""Setup."""

from decimal import Decimal
from distutils.cmd import Command
from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import sys


VERSION = '3.0.1'


# From https://circleci.com/blog/continuously-deploying-python-packages-to-pypi-with-circleci/
class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')
        tag_formatted_version = f'v{VERSION}'

        if tag != tag_formatted_version:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


class CoverageRatchetCommand(Command):
    description = 'Run coverage ratchet'
    user_options = []  # type: ignore
    coverage_file: str
    coverage_source_file: str
    coverage_url: str
    type_of_coverage: str

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        """Run command."""
        import xml.etree.ElementTree as ET

        tree = ET.parse(self.coverage_source_file)
        new_coverage = Decimal(tree.getroot().attrib["line-rate"]) * 100

        if not os.path.exists(self.coverage_file):
            with open(self.coverage_file, 'w') as f:
                f.write('0')

        with open(self.coverage_file, 'r') as f:
            high_water_mark = Decimal(f.read())

        if new_coverage < high_water_mark:
            raise Exception(
                f"{self.type_of_coverage} coverage used to be {high_water_mark}; "
                f"down to {new_coverage}%.  Fix by viewing '{self.coverage_url}'")
        elif new_coverage > high_water_mark:
            with open(self.coverage_file, 'w') as f:
                f.write(str(new_coverage))
            print(f"Just ratcheted coverage up to {new_coverage}%")
        else:
            print(f"Code coverage steady at {new_coverage}%")


class TestCoverageRatchetCommand(CoverageRatchetCommand):
    def initialize_options(self) -> None:
        """Set default values for options."""
        self.type_of_coverage = 'Test'
        self.coverage_url = 'cover/index.html'
        self.coverage_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'metrics',
            'coverage_high_water_mark'
        )
        self.coverage_source_file = "coverage.xml"


class MypyCoverageRatchetCommand(CoverageRatchetCommand):
    def initialize_options(self) -> None:
        """Set default values for options."""
        self.type_of_coverage = 'Mypy'
        self.coverage_url = 'typecover/index.html'
        self.coverage_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'metrics',
            'mypy_high_water_mark'
        )
        self.coverage_source_file = "typecover/cobertura.xml"


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='db_facts',
      version=VERSION,
      description='Database connection configuration manager',
      long_description=long_description,
      download_url=f'https://github.com/bluelabsio/db-facts/tarball/{VERSION}',
      author='Vince Broz',
      author_email='opensource@bluelabs.com',
      packages=find_packages(),
      package_data={"db_facts": ["py.typed"]},
      install_requires=['jinja2', 'pyyaml'],
      entry_points={
        'console_scripts': [
            'db_facts = db_facts.__main__:main'
        ],
        "db_facts.jinja_contexts": [
            'base64 = db_facts.base64_jinja_context:pull_base64_jinja_context',
            'env = db_facts.env_jinja_context:pull_env_jinja_context',
        ]
      },
      cmdclass={
          'coverage_ratchet': TestCoverageRatchetCommand,
          'mypy_ratchet': MypyCoverageRatchetCommand,
          'verify': VerifyVersionCommand,
      },
      scripts=['bin/db-facts'],
      license='Apache Software License',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Database :: Front-Ends',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ],
      )



from .config import load_config


def format_db_description(db_name, db_description):
    if db_description is not None:
        return f"{db_name} ({db_description})"
    else:
        return db_name


def list_db_names() -> None:
    dbcli_config = load_config()

    dbs = dbcli_config.get('dbs', {})
    db_descriptions = {
        db_name: db_config.get('description', None)
        for db_name, db_config
        in dbs.items()
    }

    output = [
        format_db_description(db_name, db_description)
        for db_name, db_description
        in db_descriptions.items()
    ]
    sorted_output = sorted(list(output))
    print("Available db_names:")
    print("* ", end='')
    print("\n* ".join(sorted_output))
